"""Main script for end-to-end task querying."""

import sys
from importlib.resources import files

import hydra
import hydra.core
import hydra.core.hydra_config
from loguru import logger
from omegaconf import DictConfig, OmegaConf

config_yaml = files("aces").joinpath("configs/aces.yaml")
if not config_yaml.is_file():
    raise FileNotFoundError("Core configuration not successfully installed!")

if len(sys.argv) == 1:
    print("Usage: aces-cli [OPTIONS]")
    print("Try 'aces-cli --help' for more information.")
    print("For more information, visit: https://eventstreamaces.readthedocs.io/en/latest/usage.html")
    sys.exit(1)


@hydra.main(version_base=None, config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig):
    import os
    from datetime import datetime
    from pathlib import Path

    from . import config, predicates, query, utils

    utils.hydra_loguru_init(f"{hydra.core.hydra_config.HydraConfig.get().job.name}.log")

    st = datetime.now()

    # Set output path
    cohort_dir = Path(cfg.cohort_dir)
    cohort_dir.mkdir(exist_ok=True, parents=True)

    # load configuration
    logger.info(f"Loading config from '{cfg.config_path}'")
    task_cfg = config.TaskExtractorConfig.load(Path(cfg.config_path))

    logger.info(f"Attempting to get predicates dataframe given:\n{OmegaConf.to_yaml(cfg.data)}")
    predicates_df = predicates.get_predicates_df(task_cfg, cfg.data)

    # query results
    result = query.query(task_cfg, predicates_df)

    # save results to parquet
    os.makedirs(os.path.dirname(cfg.output_filepath), exist_ok=True)
    result.write_parquet(cfg.output_filepath, use_pyarrow=True)
    logger.info(f"Completed in {datetime.now() - st}. Results saved to '{cfg.output_filepath}'.")


if __name__ == "__main__":
    main()



"""Tests the full end-to-end extraction process."""

import rootutils

root = rootutils.setup_root(__file__, dotenv=True, pythonpath=True, cwd=True)

import subprocess
import tempfile
from pathlib import Path

import polars as pl
from loguru import logger
from polars.testing import assert_frame_equal

pl.enable_string_cache()

TS_FORMAT = "%m/%d/%Y %H:%M"
PRED_CNT_TYPE = pl.Int64
ANY_EVENT_COLUMN = "_ANY_EVENT"

# Data (input)
PREDICATES_CSV = """
subject_id,timestamp,admission,death,discharge,lab,spo2,normal_spo2,abnormally_low_spo2,abnormally_high_spo2,procedure_start,procedure_end,ventilation,diagnosis_ICD9CM/41071,diagnosis_ICD10CM/I214
1,12/1/1989 12:03,1,0,0,0,0,0,0,0,0,0,0,0,0
1,12/1/1989 13:14,0,0,0,1,1,1,0,0,0,0,0,0,0
1,12/1/1989 15:17,0,0,0,1,1,1,0,0,0,0,0,0,0
1,12/1/1989 16:17,0,0,0,1,1,1,0,0,0,0,0,0,0
1,12/1/1989 20:17,0,0,0,1,1,1,0,0,0,0,0,0,0
1,12/2/1989 3:00,0,0,0,1,1,1,0,0,0,0,0,0,0
1,12/2/1989 9:00,0,0,0,0,0,0,0,0,0,0,0,0,0
1,12/2/1989 10:00,0,0,0,0,0,0,0,0,1,0,1,0,0
1,12/2/1989 14:22,0,0,0,0,0,0,0,0,0,1,1,0,0
1,12/2/1989 15:00,0,0,1,0,0,0,0,0,0,0,0,0,0
1,1/21/1991 11:59,0,0,0,0,0,0,0,0,0,0,0,1,0
1,1/27/1991 23:32,1,0,0,0,0,0,0,0,0,0,0,0,0
1,1/27/1991 23:46,0,0,0,1,1,1,0,0,0,0,0,0,0
1,1/28/1991 3:18,0,0,0,1,0,0,0,0,0,0,0,0,0
1,1/28/1991 3:28,0,0,0,0,0,0,0,0,1,0,1,0,0
1,1/28/1991 4:36,0,0,0,1,1,1,0,0,0,0,0,0,0
1,1/29/1991 23:32,0,0,0,1,1,1,0,0,0,0,0,0,0
1,1/30/1991 5:00,0,0,0,1,0,0,0,0,0,0,0,0,0
1,1/30/1991 8:00,0,0,0,1,1,0,0,1,0,0,0,0,0
1,1/30/1991 11:00,0,0,0,1,1,1,0,0,0,0,0,0,0
1,1/30/1991 14:00,0,0,0,0,0,0,0,0,0,0,0,0,0
1,1/30/1991 14:15,0,0,0,1,0,0,0,0,0,0,0,0,0
1,1/31/1991 1:00,0,0,0,0,0,0,0,0,0,1,1,0,0
1,1/31/1991 2:15,0,0,1,0,0,0,0,0,0,0,0,0,0
1,2/8/1991 8:15,0,0,0,1,1,1,0,0,0,0,0,0,0
1,3/3/1991 19:33,1,0,0,0,0,0,0,0,0,0,0,0,0
1,3/3/1991 20:33,0,0,0,1,1,0,1,0,0,0,0,0,0
1,3/3/1991 21:38,0,1,0,0,0,0,0,0,0,0,0,0,0
2,3/8/1996 2:24,1,0,0,0,0,0,0,0,0,0,0,0,0
2,3/8/1996 2:35,0,0,0,1,1,1,0,0,0,0,0,0,0
2,3/8/1996 4:00,0,0,0,1,1,1,0,0,0,0,0,0,0
2,3/8/1996 10:00,0,0,0,0,0,0,0,0,0,0,0,0,0
2,3/8/1996 16:00,0,0,1,0,0,0,0,0,0,0,0,0,0
2,6/5/1996 0:32,1,0,0,0,0,0,0,0,0,0,0,0,0
2,6/5/1996 0:48,0,0,0,0,0,0,0,0,0,0,0,0,1
2,6/5/1996 1:59,0,0,0,0,0,0,0,0,1,0,1,0,0
2,6/7/1996 6:00,0,0,0,1,0,0,0,0,0,0,0,0,0
2,6/7/1996 9:00,0,0,0,1,1,0,1,0,0,0,0,0,0
2,6/7/1996 12:00,0,0,0,1,1,1,0,0,0,0,0,0,0
2,6/7/1996 15:00,0,0,0,0,0,0,0,0,0,1,1,0,0
2,6/7/1996 15:00,0,0,0,0,0,0,0,0,0,0,0,0,0
2,6/8/1996 3:00,0,1,0,0,0,0,0,0,0,0,0,0,0
3,3/8/1996 2:22,0,0,0,0,0,0,0,0,1,0,1,0,0
3,3/8/1996 2:24,1,0,0,0,0,0,0,0,0,0,0,0,0
3,3/8/1996 2:37,0,0,0,1,1,1,0,0,0,0,0,0,0
3,3/9/1996 8:00,0,0,0,1,0,0,0,0,0,0,0,0,0
3,3/9/1996 11:00,0,0,0,1,1,1,0,0,0,0,0,0,0
3,3/9/1996 19:00,0,0,0,1,1,1,0,0,0,0,0,0,0
3,3/9/1996 22:00,0,0,0,0,0,0,0,0,0,0,0,0,0
3,3/11/1996 21:00,0,0,0,0,0,0,0,0,0,1,1,0,0
3,3/12/1996 0:00,0,1,0,0,0,0,0,0,0,0,0,0,0
"""

# Tasks (input)
TASKS_CFGS = {
    "inhospital_mortality": """
      # Task: 24-hour In-hospital Mortality Prediction
      predicates:
        admission:
          code: event_type//ADMISSION
        discharge:
          code: event_type//DISCHARGE
        death:
          code: event_type//DEATH
        discharge_or_death:
          expr: or(discharge, death)

      trigger: admission

      windows:
        input:
          start: NULL
          end: trigger + 24h
          start_inclusive: True
          end_inclusive: True
          has:
            _ANY_EVENT: (5, None)
          index_timestamp: end
        gap:
          start: trigger
          end: start + 48h
          start_inclusive: False
          end_inclusive: True
          has:
            admission: (None, 0)
            discharge: (None, 0)
            death: (None, 0)
        target:
          start: gap.end
          end: start -> discharge_or_death
          start_inclusive: False
          end_inclusive: True
          label: death
      """
}

# Expected output
EXPECTED_OUTPUT = {
    "inhospital_mortality": {
        "subject_id": [1, 2],
        "index_timestamp": ["01/28/1991 23:32", "06/06/1996 00:32"],
        "label": [0, 1],
        "trigger": ["01/27/1991 23:32", "06/05/1996 00:32"],
        "input.end_summary": [
            {
                "window_name": "input.end",
                "timestamp_at_start": "01/27/1991 23:32",
                "timestamp_at_end": "01/28/1991 23:32",
                "admission": 0,
                "discharge": 0,
                "death": 0,
                "discharge_or_death": 0,
                "_ANY_EVENT": 4,
            },
            {
                "window_name": "input.end",
                "timestamp_at_start": "06/05/1996 00:32",
                "timestamp_at_end": "06/06/1996 00:32",
                "admission": 0,
                "discharge": 0,
                "death": 0,
                "discharge_or_death": 0,
                "_ANY_EVENT": 2,
            },
        ],
        "input.start_summary": [
            {
                "window_name": "input.start",
                "timestamp_at_start": "12/01/1989 12:03",
                "timestamp_at_end": "01/28/1991 23:32",
                "admission": 2,
                "discharge": 1,
                "death": 0,
                "discharge_or_death": 1,
                "_ANY_EVENT": 16,
            },
            {
                "window_name": "input.start",
                "timestamp_at_start": "03/08/1996 02:24",
                "timestamp_at_end": "06/06/1996 00:32",
                "admission": 2,
                "discharge": 1,
                "death": 0,
                "discharge_or_death": 1,
                "_ANY_EVENT": 8,
            },
        ],
        "gap.end_summary": [
            {
                "window_name": "gap.end",
                "timestamp_at_start": "01/27/1991 23:32",
                "timestamp_at_end": "01/29/1991 23:32",
                "admission": 0,
                "discharge": 0,
                "death": 0,
                "discharge_or_death": 0,
                "_ANY_EVENT": 5,
            },
            {
                "window_name": "gap.end",
                "timestamp_at_start": "06/05/1996 00:32",
                "timestamp_at_end": "06/07/1996 00:32",
                "admission": 0,
                "discharge": 0,
                "death": 0,
                "discharge_or_death": 0,
                "_ANY_EVENT": 2,
            },
        ],
        "target.end_summary": [
            {
                "window_name": "target.end",
                "timestamp_at_start": "01/29/1991 23:32",
                "timestamp_at_end": "01/31/1991 02:15",
                "admission": 0,
                "discharge": 1,
                "death": 0,
                "discharge_or_death": 1,
                "_ANY_EVENT": 7,
            },
            {
                "window_name": "target.end",
                "timestamp_at_start": "06/07/1996 00:32",
                "timestamp_at_end": "06/08/1996 03:00",
                "admission": 0,
                "discharge": 0,
                "death": 1,
                "discharge_or_death": 1,
                "_ANY_EVENT": 5,
            },
        ],
    }
}


def run_command(script: str, hydra_kwargs: dict[str, str], test_name: str):
    command_parts = [script] + [f"{k}={v}" for k, v in hydra_kwargs.items()]
    command_out = subprocess.run(" ".join(command_parts), shell=True, capture_output=True)
    stderr = command_out.stderr.decode()
    stdout = command_out.stdout.decode()
    if command_out.returncode != 0:
        raise AssertionError(f"{test_name} failed!\nstdout:\n{stdout}\nstderr:\n{stderr}")
    return stderr, stdout


def assert_df_equal(want: pl.DataFrame, got: pl.DataFrame, msg: str = None, **kwargs):
    try:
        assert_frame_equal(want, got, **kwargs)
    except AssertionError as e:
        pl.Config.set_tbl_rows(-1)
        print(f"DFs are not equal: {msg}\nWant:")
        print(want)
        print("Got:")
        print(got)
        raise AssertionError(f"{msg}\n{e}") from e


def test_e2e():
    with tempfile.TemporaryDirectory() as d:
        data_dir = Path(d) / "sample_data"
        configs_dir = Path(d) / "sample_configs"
        output_dir = Path(d) / "sample_output"

        # Create the directories
        data_dir.mkdir()
        configs_dir.mkdir()
        output_dir.mkdir()

        # Write the predicates CSV file
        predicates_csv = data_dir / "sample_data.csv"
        predicates_csv.write_text(PREDICATES_CSV.strip())

        # Run script and check the outputs
        all_stderrs = []
        all_stdouts = []
        full_stderr = ""
        full_stdout = ""
        try:
            for task_name, task_cfg in TASKS_CFGS.items():
                logger.info(f"Running task '{task_name}'...")

                # Write the task config YAMLs
                task_cfg_path = configs_dir / f"{task_name}.yaml"
                task_cfg_path.write_text(task_cfg)

                output_path = output_dir / f"{task_name}.parquet"

                extraction_config_kwargs = {
                    "data.path": str(predicates_csv.resolve()),
                    "data.standard": "direct",
                    "cohort_dir": str(configs_dir.resolve()),
                    "cohort_name": task_name,
                    "output_filepath": str(output_path.resolve()),
                    "hydra.verbose": True,
                }

                stderr, stdout = run_command("aces-cli", extraction_config_kwargs, task_name)
                stderr, stdout = run_command("aces-cli", extraction_config_kwargs, task_name)

                all_stderrs.append(stderr)
                all_stdouts.append(stdout)

                full_stderr = "\n".join(all_stderrs)
                full_stdout = "\n".join(all_stdouts)

                fp = output_dir / f"{task_name}.parquet"
                assert fp.is_file(), f"Expected {fp} to exist."
                got_df = pl.read_parquet(fp)

                # Check the columns
                expected_columns = EXPECTED_OUTPUT[task_name].keys()
                assert got_df.columns == list(expected_columns), f"Columns mismatch for task '{task_name}'"

                # Check the data
                for col_name, expected_data in EXPECTED_OUTPUT[task_name].items():
                    if col_name in ["index_timestamp", "trigger"]:
                        want = pl.DataFrame({col_name: expected_data}).with_columns(
                            pl.col(col_name).str.strptime(pl.Datetime, format=TS_FORMAT)
                        )
                    elif col_name.endswith("_summary"):
                        df_struct = pl.DataFrame(expected_data)
                        df_struct = df_struct.with_columns(
                            pl.col("timestamp_at_start").str.strptime(pl.Datetime, format=TS_FORMAT),
                            pl.col("timestamp_at_end").str.strptime(pl.Datetime, format=TS_FORMAT),
                        )
                        want = df_struct.select(
                            pl.struct(*[col for col in df_struct.columns]).alias(col_name)
                        )
                    else:
                        want = pl.DataFrame({col_name: expected_data}).with_columns(
                            pl.col(col_name).cast(PRED_CNT_TYPE)
                        )
                    got = got_df.select(col_name)
                    assert_df_equal(want, got, f"Data mismatch for task '{task_name}', column '{col_name}'")

        except AssertionError as e:
            print(f"Failed on task '{task_name}'")
            print(f"stderr:\n{full_stderr}")
            print(f"stdout:\n{full_stdout}")
            raise e



#!/usr/bin/env python

import sys


def main():
    result = []
    for arg in sys.argv[1:]:
        prefix = arg[: arg.rfind("/")]
        num = int(arg[arg.rfind("/") + 1 :])
        result.extend(f"{prefix}/{i}" for i in range(num))

    print(",".join(result))


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
import os
import shutil
import sys
from pathlib import Path

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Path setup

__location__ = Path(os.path.dirname(__file__))
__src__ = __location__ / "../.."

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, str(__src__))


# -- Project information

project = "Event Stream Automatic Cohort Extraction System (ACES)"
copyright = "2024, Justin Xu & Matthew McDermott"
author = "Justin Xu & Matthew McDermott"

release = "0.2.2"
version = "0.2.2"


def ensure_pandoc_installed(_):
    """Source: https://stackoverflow.com/questions/62398231/building-docs-fails-due-to-missing-pandoc"""
    import pypandoc

    # Download pandoc if necessary. If pandoc is already installed and on
    # the PATH, the installed version will be used. Otherwise, we will
    # download a copy of pandoc into docs/bin/ and add that to our PATH.
    pandoc_dir = str(__location__ / "bin")
    # Add dir containing pandoc binary to the PATH environment variable
    if pandoc_dir not in os.environ["PATH"].split(os.pathsep):
        os.environ["PATH"] += os.pathsep + pandoc_dir

    pypandoc.ensure_pandoc_installed(
        targetfolder=pandoc_dir,
        delete_installer=True,
    )


# -- Run sphinx-apidoc
# This ensures we don't need to run apidoc manually.

# TODO: use https://github.com/sphinx-extensions2/sphinx-autodoc2

from sphinx.ext import apidoc

output_dir = __location__ / "api"
module_dir = __src__ / "src/aces"
if output_dir.is_dir():
    shutil.rmtree(output_dir)

try:
    cmd_line = f"--implicit-namespaces -e -f -o {output_dir} {module_dir}"
    apidoc.main(cmd_line.split(" "))
except Exception as e:  # pylint: disable=broad-except
    print(f"Running `sphinx-apidoc {cmd_line}` failed!\n{e}")


# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinxcontrib.collections",
    "sphinx_subfigure",
    "sphinx_immaterial",
    "myst_parser",
    "nbsphinx",
]

nbsphinx_allow_errors = True


collections_dir = __location__ / "_collections"
if not collections_dir.is_dir():
    os.mkdir(collections_dir)

python_version = ".".join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "python": ("https://docs.python.org/" + python_version, None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "pandera": ("https://pandera.readthedocs.io/en/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "setuptools": ("https://setuptools.pypa.io/en/stable/", None),
    "pyscaffold": ("https://pyscaffold.org/en/stable", None),
    "hyperimpute": ("https://hyperimpute.readthedocs.io/en/latest/", None),
    "xgbse": ("https://loft-br.github.io/xgboost-survival-embeddings/", None),
    "lifelines": ("https://lifelines.readthedocs.io/en/stable/", None),
    "optuna": ("https://optuna.readthedocs.io/en/stable/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Control options for included jupyter notebooks.
nb_execution_mode = "off"


# -- Options for HTML output

# Configure MyST-Parser
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "substitution",
    "tasklist",
]

myst_update_mathjax = True

# MyST URL schemes.
myst_url_schemes = {
    "http": None,
    "https": None,
    "ftp": None,
    "mailto": None,
    "repo-code": "https://github.com/justin13601/ACES/tree/main/{{path}}#{{fragment}}",
    # "doi": "https://doi.org/{{path}}",
    # "gh-issue": {
    #     "url": "https://github.com/executablebooks/MyST-Parser/issue/{{path}}#{{fragment}}",
    #     "title": "Issue #{{path}}",
    #     "classes": ["github"],
    # },
}

# The suffix of source filenames.
source_suffix = [".rst", ".md"]

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "py:obj"

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# https://pygments.org/styles/
pygments_style = "tango"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If this is True, todo emits a warning for each TODO entries. The default is False.
todo_emit_warnings = True


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_immaterial"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

# Material theme options (see theme.conf for more information)
html_theme_options = {
    # Set the name of the project to appear in the navigation.
    "nav_title": "ACES",
    # Icon,
    # "logo_icon": "query-512.png",
    # Set you GA account ID to enable tracking
    # "google_analytics_account": "UA-XXXXX",
    #
    # TODO: Sitemap.
    # Specify a base_url used to generate sitemap.xml. If not
    # specified, then no sitemap will be built.
    # 'base_url': 'https://project.github.io/project',
    #
    # Set the color and the accent color
    "palette": {"primary": "green", "accent": "green"},
    # {
    #     "media": "(prefers-color-scheme: light)",
    #     "scheme": "default",
    #     "toggle": {
    #         "icon": "material/toggle-switch-off-outline",
    #         "name": "Switch to dark mode",
    #     },
    # },
    # {
    #     "media": "(prefers-color-scheme: dark)",
    #     "scheme": "slate",
    #     "toggle": {
    #         "icon": "material/toggle-switch",
    #         "name": "Switch to light mode",
    #     },
    # },
    # "color_primary": "green",
    # "color_accent": "green",
    # Set the repo location to get a badge with stats
    "repo_url": "https://github.com/justin13601/ACES",
    "repo_name": "ACES",
    # Visible levels of the global TOC; -1 means unlimited
    "globaltoc_depth": 3,
    # If False, expand all TOC entries
    "globaltoc_collapse": True,
    # If True, show hidden TOC entries
    "globaltoc_includehidden": False,
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = f"ACES v{version} Documentation"

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "ACES Documentation"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "assets/aces_logo.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "assets/aces_logo.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {"**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "aces-doc"


# -- Options for LaTeX output

latex_elements = {  # type: ignore
    # The paper size ("letterpaper" or "a4paper").
    # "papersize": "letterpaper",
    # The font size ("10pt", "11pt" or "12pt").
    # "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    # "preamble": "",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", "user_guide.tex", "ACES Documentation", "Justin Xu & Matthew McDermott", "manual")
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = ""

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True

# -- Options for EPUB output
epub_show_urls = "footnote"

print(f"loading configurations for {project} {version} ...", file=sys.stderr)


def setup(app):
    app.connect("builder-inited", ensure_pandoc_installed)



"""This module contains the main function for querying a task.

It accepts the configuration file and predicate columns, builds the tree, and recursively queries the tree.
"""


import polars as pl
from bigtree import preorder_iter
from loguru import logger

from .config import TaskExtractorConfig
from .constraints import check_constraints
from .extract_subtree import extract_subtree
from .utils import log_tree


def query(cfg: TaskExtractorConfig, predicates_df: pl.DataFrame) -> pl.DataFrame:
    """Query a task using the provided configuration file and predicates dataframe.

    Args:
        cfg: TaskExtractorConfig object of the configuration file.
        predicates_df: Polars predicates dataframe.

    Returns:
        polars.DataFrame: The result of the task query, containing subjects who satisfy the conditions
        defined in cfg. Timestamps for the start/end boundaries of each window specified in the task
        configuration, as well as predicate counts for each window, are provided.
    """
    if not isinstance(predicates_df, pl.DataFrame):
        raise TypeError(f"Predicates dataframe type must be a polars.DataFrame. Got: {type(predicates_df)}.")

    logger.info("Checking if '(subject_id, timestamp)' columns are unique...")
    try:
        assert (
            predicates_df.n_unique(subset=["subject_id", "timestamp"]) == predicates_df.shape[0]
        ), "The (subject_id, timestamp) columns must be unique."
    except AssertionError as e:
        logger.error(str(e))
        return pl.DataFrame()

    log_tree(cfg.window_tree)

    logger.info("Beginning query...")
    logger.info("Identifying possible trigger nodes based on the specified trigger event...")
    prospective_root_anchors = check_constraints({cfg.trigger.predicate: (1, None)}, predicates_df).select(
        "subject_id", pl.col("timestamp").alias("subtree_anchor_timestamp")
    )
    try:
        assert (
            not prospective_root_anchors.is_empty()
        ), f"No valid rows found for the trigger event '{cfg.trigger.predicate}'. Exiting."
    except AssertionError as e:
        logger.error(str(e))
        return pl.DataFrame()

    result = extract_subtree(cfg.window_tree, prospective_root_anchors, predicates_df)
    if result.is_empty():
        logger.info("No valid rows found.")
    else:
        logger.info(f"Done. {result.shape[0]:,} valid rows returned.")

    result = result.rename({"subtree_anchor_timestamp": "trigger"})

    to_return_cols = [
        "subject_id",
        "trigger",
        *[f"{node.node_name}_summary" for node in preorder_iter(cfg.window_tree)][1:],
    ]

    # add label column if specified
    if cfg.label_window:
        logger.info(
            f"Extracting label '{cfg.windows[cfg.label_window].label}' from window "
            f"'{cfg.label_window}'..."
        )
        label_col = "end" if cfg.windows[cfg.label_window].root_node == "start" else "start"
        result = result.with_columns(
            pl.col(f"{cfg.label_window}.{label_col}_summary")
            .struct.field(cfg.windows[cfg.label_window].label)
            .alias("label")
        )
        to_return_cols.insert(1, "label")

    # add index_timestamp column if specified
    if cfg.index_timestamp_window:
        logger.info(
            f"Setting index timestamp as '{cfg.windows[cfg.index_timestamp_window].index_timestamp}' "
            f"of window '{cfg.index_timestamp_window}'..."
        )
        index_timestamp_col = (
            "end" if cfg.windows[cfg.index_timestamp_window].root_node == "start" else "start"
        )
        result = result.with_columns(
            pl.col(f"{cfg.index_timestamp_window}.{index_timestamp_col}_summary")
            .struct.field(f"timestamp_at_{cfg.windows[cfg.index_timestamp_window].index_timestamp}")
            .alias("index_timestamp")
        )
        to_return_cols.insert(1, "index_timestamp")

    return result.select(to_return_cols)



"""This module contains the functions for extracting constraint hierarchy subtrees."""

from datetime import timedelta

import polars as pl
from bigtree import Node
from loguru import logger

from .aggregate import aggregate_event_bound_window, aggregate_temporal_window
from .constraints import check_constraints


def extract_subtree(
    subtree: Node,
    subtree_anchor_realizations: pl.DataFrame,
    predicates_df: pl.DataFrame,
    subtree_root_offset: timedelta = timedelta(0),
) -> pl.DataFrame:
    """The main algorithmic recursive call to identify valid realizations of a subtree.

    This function takes in a global ``predicates_df``, a subtree of constraints, and the temporal offset that
    any realization the root timestamp of the subtree would have relative to the corresponding subtree anchor.
    It will use this information to recurse through the subtree and identify any valid realizations of this
    subtree, returning them in a dataframe keyed by the subtree anchor event timestamps and with a series of
    columns containing subtree edge start and end timestamps and contained predicate counts.

    Args:
        subtree: The subtree to extract realizations from. This is specified through a `BigTree.Node` object.
            This ``Node`` object can have zero or more children, each of which must have the following:
              - ``name``: The name of the subtree root.
              - ``constraints``: The constraints associated with the subtree root, structured as a dictionary
                from predicate column name to a tuple containing the valid (inclusive) minimum and maximum
                values the predicate counts can take on (use `None` for no constraint).
              - ``endpoint_expr``: A tuple containing the endpoint expression for the subtree root. This
                should be either a `ToEventWindowBounds` or a `TemporalWindowBounds` formatted tuple object,
                less the offset parameter, as that is something determined by the structure of the tree, not
                pre-set in the configuration.
        subtree_anchor_realizations: The dataframe containing the anchor to subtree root mapping. This
            dataframe will have the following columns:
              - ``"subject_id"``: The ID of the subject. All analyses will be performed within ``subject_id``
                groups.
              - ``subtree_anchor_timestamp``: The timestamp of all possible prospective subtree anchor
                realizations. These will all correspond to extant events (``subject_id``, ``timestamp`` pairs
                in ``predicates_df``).
        predicates_df: The dataframe containing the predicates to summarize. This dataframe will have the
            following mandatory columns:
        subtree_root_offset: The temporal offset of the subtree root relative to the subtree anchor.

    Returns:
        pl.DataFrame: The result of the subtree extraction, containing subjects who satisfy the conditions
        defined in the subtree. Timestamps for the start/end boundaries of each window specified in the
        subtree configuration, as well as predicate counts for each window, are provided.

    Examples:
        >>> from bigtree import Node
        >>> from datetime import datetime
        >>> # We'll use an example for in-hospital mortality prediction. Our root event of the tree will be
        >>> # an admission event.
        >>> root = Node("admission")
        >>> #
        >>> #### BRANCH 1 ####
        >>> # Our first branch off of admission will be checking a gap window, then our target window.
        >>> # Node 1 will represent our gap window. We say that in the 24 hours after the admission, there
        >>> # should be no discharges, deaths, or covid events.
        >>> gap_node = Node("gap") # This sets the node's name.
        >>> gap_node.endpoint_expr = (True, timedelta(days=2), True)
        >>> gap_node.constraints = {
        ...     "is_discharge": (None, 0), "is_death": (None, 0), "is_covid_dx": (None, 0)
        ... }
        >>> gap_node.parent = root
        >>> # Node 2 will start our target window and span until the next discharge or death event.
        >>> # There should be no covid events.
        >>> target_node = Node("target") # This sets the node's name.
        >>> target_node.endpoint_expr = (True, "is_discharge", True)
        >>> target_node.constraints = {"is_covid_dx": (None, 0)}
        >>> target_node.parent = gap_node
        >>> #
        >>> #### BRANCH 2 ####
        >>> # Finally, for our second branch, we will impose no constraints but track the input time range,
        >>> # which will span from the beginning of the record to 24 hours after admission.
        >>> input_end_node = Node("input_end")
        >>> input_end_node.endpoint_expr = (True, timedelta(days=1), True)
        >>> input_end_node.constraints = {}
        >>> input_end_node.parent = root
        >>> input_start_node = Node("input_start")
        >>> input_start_node.endpoint_expr = (True, "-_RECORD_START", True)
        >>> input_start_node.constraints = {}
        >>> input_start_node.parent = root
        >>> #
        >>> #### BRANCH 3 ####
        >>> # For our last branch, we will validate that the patient has sufficient historical data, asserting
        >>> # that they should have at least 1 event of any kind at least 1 year prior to the trigger event.
        >>> # This will be expressed through two windows, one spanning back a year, and the other looking
        >>> # prior to that year.
        >>> pre_node_1yr = Node("pre_node_1yr")
        >>> pre_node_1yr.endpoint_expr = (False, timedelta(days=-365), False)
        >>> pre_node_1yr.constraints = {}
        >>> pre_node_1yr.parent = root
        >>> pre_node_total = Node("pre_node_total")
        >>> pre_node_total.endpoint_expr = (False, "-_RECORD_START", False)
        >>> pre_node_total.constraints = {"*": (1, None)}
        >>> pre_node_total.parent = pre_node_1yr
        >>> #
        >>> #### PREDICATES_DF ####
        >>> # We'll have the following patient data:
        >>> #  - subject 1 will have an admission that won't count because they'll have a covid diagnosis,
        >>> #    then an admission that won't count because there will be no associated discharge.
        >>> #  - subject 2 will have an admission that won't count because they'll have too little data before
        >>> #    it, then a second admission that will count.
        >>> #  - subject 3 will have an admission that will be too short.
        >>> #
        >>> predicates_df = pl.DataFrame({
        ...     "subject_id": [
        ...         1, 1, 1, 1, 1, # Pre-event, Admission, Covid, Discharge, Admission.
        ...         2, 2, 2, 2, 2, # Pre-event-too-close, Admission, Discharge, Admission, Death & Discharge.
        ...         3, 3, 3,       # Pre-event, Admission, Death
        ...     ],
        ...     "timestamp": [
        ...         # Subject 1
        ...         datetime(year=1980, month=12, day=1,  hour=12, minute=3),  # Pre-event
        ...         datetime(year=1989, month=12, day=3,  hour=13, minute=14), # Admission
        ...         datetime(year=1989, month=12, day=5,  hour=15, minute=17), # Covid
        ...         datetime(year=1989, month=12, day=7,  hour=11, minute=4),  # Discharge
        ...         datetime(year=1989, month=12, day=23, hour=3,  minute=12), # Admission
        ...         # Subject 2
        ...         datetime(year=1983, month=12, day=1,  hour=22, minute=2),  # Pre-event-too-close
        ...         datetime(year=1983, month=12, day=2,  hour=12, minute=3),  # Admission
        ...         datetime(year=1983, month=12, day=8,  hour=13, minute=14), # Discharge
        ...         datetime(year=1989, month=12, day=6,  hour=15, minute=17), # Valid Admission
        ...         datetime(year=1989, month=12, day=10, hour=16, minute=22), # Death & Discharge
        ...         # Subject 3
        ...         datetime(year=1982, month=2,  day=13, hour=10, minute=44), # Pre-event
        ...         datetime(year=1999, month=12, day=6,  hour=15, minute=17), # Admission
        ...         datetime(year=1999, month=12, day=6,  hour=16, minute=22), # Discharge
        ...     ],
        ...     "is_admission": [0, 1, 0, 0, 1,   0, 1, 0, 1, 0,   0, 1, 0],
        ...     "is_discharge": [0, 0, 0, 1, 0,   0, 0, 1, 0, 1,   0, 0, 1],
        ...     "is_death":     [0, 0, 0, 0, 0,   0, 0, 0, 0, 1,   0, 0, 0],
        ...     "is_covid_dx":  [0, 0, 1, 0, 0,   0, 0, 0, 0, 0,   0, 0, 0],
        ...     "_ANY_EVENT":   [1, 1, 1, 1, 1,   1, 1, 1, 1, 1,   1, 1, 1],
        ... })
        >>> subtreee_anchor_realizations = (
        ...     predicates_df.filter(pl.col("is_admission") > 0)
        ...     .rename({"timestamp": "subtree_anchor_timestamp"})
        ... ).select("subject_id", "subtree_anchor_timestamp")
        >>> print(subtreee_anchor_realizations)
        shape: (5, 2)
        ┌────────────┬──────────────────────────┐
        │ subject_id ┆ subtree_anchor_timestamp │
        │ ---        ┆ ---                      │
        │ i64        ┆ datetime[μs]             │
        ╞════════════╪══════════════════════════╡
        │ 1          ┆ 1989-12-03 13:14:00      │
        │ 1          ┆ 1989-12-23 03:12:00      │
        │ 2          ┆ 1983-12-02 12:03:00      │
        │ 2          ┆ 1989-12-06 15:17:00      │
        │ 3          ┆ 1999-12-06 15:17:00      │
        └────────────┴──────────────────────────┘
        >>> out = extract_subtree(root, subtreee_anchor_realizations, predicates_df, timedelta(0))
        >>> out.select(
        ...     "subject_id",
        ...     "subtree_anchor_timestamp",
        ... )
        shape: (1, 2)
        ┌────────────┬──────────────────────────┐
        │ subject_id ┆ subtree_anchor_timestamp │
        │ ---        ┆ ---                      │
        │ i64        ┆ datetime[μs]             │
        ╞════════════╪══════════════════════════╡
        │ 2          ┆ 1989-12-06 15:17:00      │
        └────────────┴──────────────────────────┘
        >>> out.columns # doctest: +NORMALIZE_WHITESPACE
        ['subject_id',
         'target_summary',
         'subtree_anchor_timestamp',
         'gap_summary',
         'input_end_summary',
         'input_start_summary',
         'pre_node_total_summary',
         'pre_node_1yr_summary']
        >>> def print_window(name: str, do_drop_any_events: bool = True):
        ...     drop_cols = ["window_name", "subject_id", "subtree_anchor_timestamp"]
        ...     if do_drop_any_events:
        ...         drop_cols.append("_ANY_EVENT")
        ...     return (
        ...         out.select("subject_id", "subtree_anchor_timestamp", name)
        ...         .unnest(name)
        ...         .drop(*drop_cols)
        ...     )
        >>> print_window("gap_summary")
        shape: (1, 6)
        ┌─────────────────────┬─────────────────────┬──────────────┬──────────────┬──────────┬─────────────┐
        │ timestamp_at_start  ┆ timestamp_at_end    ┆ is_admission ┆ is_discharge ┆ is_death ┆ is_covid_dx │
        │ ---                 ┆ ---                 ┆ ---          ┆ ---          ┆ ---      ┆ ---         │
        │ datetime[μs]        ┆ datetime[μs]        ┆ i64          ┆ i64          ┆ i64      ┆ i64         │
        ╞═════════════════════╪═════════════════════╪══════════════╪══════════════╪══════════╪═════════════╡
        │ 1989-12-06 15:17:00 ┆ 1989-12-08 15:17:00 ┆ 1            ┆ 0            ┆ 0        ┆ 0           │
        └─────────────────────┴─────────────────────┴──────────────┴──────────────┴──────────┴─────────────┘
        >>> print_window("target_summary")
        shape: (1, 6)
        ┌─────────────────────┬─────────────────────┬──────────────┬──────────────┬──────────┬─────────────┐
        │ timestamp_at_start  ┆ timestamp_at_end    ┆ is_admission ┆ is_discharge ┆ is_death ┆ is_covid_dx │
        │ ---                 ┆ ---                 ┆ ---          ┆ ---          ┆ ---      ┆ ---         │
        │ datetime[μs]        ┆ datetime[μs]        ┆ i64          ┆ i64          ┆ i64      ┆ i64         │
        ╞═════════════════════╪═════════════════════╪══════════════╪══════════════╪══════════╪═════════════╡
        │ 1989-12-08 15:17:00 ┆ 1989-12-10 16:22:00 ┆ 0            ┆ 1            ┆ 1        ┆ 0           │
        └─────────────────────┴─────────────────────┴──────────────┴──────────────┴──────────┴─────────────┘
        >>> print_window("input_start_summary")
        shape: (1, 6)
        ┌─────────────────────┬─────────────────────┬──────────────┬──────────────┬──────────┬─────────────┐
        │ timestamp_at_start  ┆ timestamp_at_end    ┆ is_admission ┆ is_discharge ┆ is_death ┆ is_covid_dx │
        │ ---                 ┆ ---                 ┆ ---          ┆ ---          ┆ ---      ┆ ---         │
        │ datetime[μs]        ┆ datetime[μs]        ┆ i64          ┆ i64          ┆ i64      ┆ i64         │
        ╞═════════════════════╪═════════════════════╪══════════════╪══════════════╪══════════╪═════════════╡
        │ 1983-12-01 22:02:00 ┆ 1989-12-06 15:17:00 ┆ 2            ┆ 1            ┆ 0        ┆ 0           │
        └─────────────────────┴─────────────────────┴──────────────┴──────────────┴──────────┴─────────────┘
        >>> print_window("input_end_summary")
        shape: (1, 6)
        ┌─────────────────────┬─────────────────────┬──────────────┬──────────────┬──────────┬─────────────┐
        │ timestamp_at_start  ┆ timestamp_at_end    ┆ is_admission ┆ is_discharge ┆ is_death ┆ is_covid_dx │
        │ ---                 ┆ ---                 ┆ ---          ┆ ---          ┆ ---      ┆ ---         │
        │ datetime[μs]        ┆ datetime[μs]        ┆ i64          ┆ i64          ┆ i64      ┆ i64         │
        ╞═════════════════════╪═════════════════════╪══════════════╪══════════════╪══════════╪═════════════╡
        │ 1989-12-06 15:17:00 ┆ 1989-12-07 15:17:00 ┆ 1            ┆ 0            ┆ 0        ┆ 0           │
        └─────────────────────┴─────────────────────┴──────────────┴──────────────┴──────────┴─────────────┘
        >>> print_window("pre_node_1yr_summary")
        shape: (1, 6)
        ┌─────────────────────┬─────────────────────┬──────────────┬──────────────┬──────────┬─────────────┐
        │ timestamp_at_start  ┆ timestamp_at_end    ┆ is_admission ┆ is_discharge ┆ is_death ┆ is_covid_dx │
        │ ---                 ┆ ---                 ┆ ---          ┆ ---          ┆ ---      ┆ ---         │
        │ datetime[μs]        ┆ datetime[μs]        ┆ i64          ┆ i64          ┆ i64      ┆ i64         │
        ╞═════════════════════╪═════════════════════╪══════════════╪══════════════╪══════════╪═════════════╡
        │ 1989-12-06 15:17:00 ┆ 1988-12-06 15:17:00 ┆ 0            ┆ 0            ┆ 0        ┆ 0           │
        └─────────────────────┴─────────────────────┴──────────────┴──────────────┴──────────┴─────────────┘
        >>> print_window("pre_node_total_summary")
        shape: (1, 6)
        ┌─────────────────────┬─────────────────────┬──────────────┬──────────────┬──────────┬─────────────┐
        │ timestamp_at_start  ┆ timestamp_at_end    ┆ is_admission ┆ is_discharge ┆ is_death ┆ is_covid_dx │
        │ ---                 ┆ ---                 ┆ ---          ┆ ---          ┆ ---      ┆ ---         │
        │ datetime[μs]        ┆ datetime[μs]        ┆ i64          ┆ i64          ┆ i64      ┆ i64         │
        ╞═════════════════════╪═════════════════════╪══════════════╪══════════════╪══════════╪═════════════╡
        │ 1983-12-01 22:02:00 ┆ 1988-12-06 15:17:00 ┆ 1            ┆ 1            ┆ 0        ┆ 0           │
        └─────────────────────┴─────────────────────┴──────────────┴──────────────┴──────────┴─────────────┘
    """
    recursive_results = []
    predicate_cols = [c for c in predicates_df.columns if c not in {"subject_id", "timestamp"}]

    if not subtree.children:
        return subtree_anchor_realizations

    for child in subtree.children:
        logger.info(f"Summarizing subtree rooted at '{child.name}'...")

        # Step 1: Summarize the window from the subtree.root to child
        endpoint_expr = child.endpoint_expr
        if type(endpoint_expr) is tuple:
            endpoint_expr = endpoint_expr + (subtree_root_offset,)
        else:
            endpoint_expr.offset += subtree_root_offset

        match endpoint_expr[1]:
            case timedelta():
                child_root_offset = subtree_root_offset + endpoint_expr[1]
                window_summary_df = (
                    aggregate_temporal_window(predicates_df, endpoint_expr)
                    .with_columns(
                        pl.col("timestamp").alias("subtree_anchor_timestamp"),
                        pl.col("timestamp").alias("child_anchor_timestamp"),
                    )
                    .drop("timestamp")
                )
            case str():
                # In an event bound case, the child root will be a proper extant event, so it will be the
                # anchor as well, and thus the child root offset should be zero.
                child_root_offset = timedelta(days=0)
                window_summary_df = (
                    aggregate_event_bound_window(predicates_df, endpoint_expr)
                    .with_columns(
                        pl.col("timestamp").alias("subtree_anchor_timestamp"),
                        pl.col("timestamp_at_end").alias("child_anchor_timestamp"),
                    )
                    .drop("timestamp")
                )
            case _:
                raise ValueError(f"Invalid endpoint expression: '{endpoint_expr}'")

        # Step 2: Filter to valid subtree anchors
        window_summary_df = window_summary_df.join(
            subtree_anchor_realizations, on=["subject_id", "subtree_anchor_timestamp"], how="inner"
        )

        # Step 3: Filter to where constraints are valid
        window_summary_df = check_constraints(child.constraints, window_summary_df)

        # Step 4: Produce child anchor realizations
        child_anchor_realizations = window_summary_df.select(
            "subject_id",
            pl.col("child_anchor_timestamp").alias("subtree_anchor_timestamp"),
        )

        # Step 5: Recurse
        recursive_result = extract_subtree(
            child,
            child_anchor_realizations,
            predicates_df,
            child_root_offset,
        )

        # Step 6: Join summaries and timestamps
        # Step 6.1: Convert recursive_result up to subtree anchor space.
        recursive_result = (
            recursive_result.rename({"subtree_anchor_timestamp": "child_anchor_timestamp"})
            .join(
                window_summary_df.select("subject_id", "subtree_anchor_timestamp", "child_anchor_timestamp"),
                on=["subject_id", "child_anchor_timestamp"],
                how="left",
            )
            .drop("child_anchor_timestamp")
        )

        # Step 6.2: Summarize the observed window statistics and timestamps for eventual return.
        for_return = window_summary_df.select(
            "subject_id",
            "subtree_anchor_timestamp",
            pl.struct(
                pl.lit(child.name).alias("window_name"),
                "timestamp_at_start",
                "timestamp_at_end",
                *predicate_cols,
            ).alias(f"{child.name}_summary"),
        )

        recursive_results.append(
            recursive_result.join(for_return, on=["subject_id", "subtree_anchor_timestamp"], how="left")
        )

    # Step 7: Join children recursive results where all children find a valid realization
    all_children = recursive_results[0]
    for df in recursive_results[1:]:
        all_children = all_children.join(df, on=["subject_id", "subtree_anchor_timestamp"], how="inner")

    return all_children
